@@ -9,7 +9,11 @@ use core::fmt;
 use core::mem::{align_of, size_of};
 use core::ops::{Drop, FnMut};
 use core::slice::from_raw_parts_mut;
+
+#[cfg(not(loom))]
 use core::sync::atomic::{AtomicBool, AtomicUsize, Ordering};
+#[cfg(loom)]
+pub use loom::sync::atomic::{AtomicBool, AtomicUsize, Ordering};
 
 use crossbeam_utils::CachePadded;
 
@@ -22,8 +26,11 @@ use crate::replica::MAX_THREADS_PER_REPLICA;
 const DEFAULT_LOG_BYTES: usize = 32 * 1024 * 1024;
 const_assert!(DEFAULT_LOG_BYTES >= 1 && (DEFAULT_LOG_BYTES & (DEFAULT_LOG_BYTES - 1) == 0));
 
-/// The maximum number of replicas that can be used against the log.
-const MAX_REPLICAS: usize = 192;
+/// The maximum number of replicas that can be registered with the log.
+#[cfg(not(loom))]
+pub const MAX_REPLICAS_PER_LOG: usize = 192;
+#[cfg(loom)] // Otherwise uses too much stack space wich crashes in loom...
+pub const MAX_REPLICAS_PER_LOG: usize = 3;
 
 /// Constant required for garbage collection. When the tail and the head are
 /// these many entries apart on the circular buffer, garbage collection will
@@ -118,7 +125,7 @@ where
     /// Required for garbage collection; since replicas make progress over the log
     /// independently, we want to make sure that we don't garbage collect operations
     /// that haven't been executed by all replicas.
-    ltails: [CachePadded<AtomicUsize>; MAX_REPLICAS],
+    ltails: [CachePadded<AtomicUsize>; MAX_REPLICAS_PER_LOG],
 
     /// Identifier that will be allocated to the next replica that registers with
     /// this Log. Also required to correctly index into ltails above.
@@ -127,7 +134,7 @@ where
     /// Array consisting of local alive masks for each registered replica. Required
     /// because replicas make independent progress over the log, so we need to
     /// track log wrap-arounds for each of them separately.
-    lmasks: [CachePadded<Cell<bool>>; MAX_REPLICAS],
+    lmasks: [CachePadded<Cell<bool>>; MAX_REPLICAS_PER_LOG],
 }
 
 impl<'a, T> fmt::Debug for Log<'a, T>
@@ -177,8 +184,6 @@ where
     /// This method also allocates memory for the log upfront. No further allocations
     /// will be performed once this method returns.
     pub fn new<'b>(bytes: usize) -> Log<'b, T> {
-        use arr_macro::arr;
-
         // Calculate the number of entries that will go into the log, and retrieve a
         // slice to it from the allocated region of memory.
         let mut num = bytes / Log::<T>::entry_size();
@@ -209,7 +214,7 @@ where
         let raw = unsafe { from_raw_parts_mut(mem as *mut Cell<Entry<T>>, num) };
 
         // Initialize all log entries by calling the default constructor.
-        for e in &mut raw[..] {
+        for e in raw.iter_mut() {
             unsafe {
                 ::core::ptr::write(
                     e,
@@ -222,22 +227,45 @@ where
             }
         }
 
-        let fls: [CachePadded<Cell<bool>>; MAX_REPLICAS] = arr![Default::default(); 192];
-        for idx in 0..MAX_REPLICAS {
-            fls[idx].set(true)
-        }
+        #[allow(clippy::declare_interior_mutable_const)]
+        const LMASK_DEFAULT: CachePadded<Cell<bool>> = CachePadded::new(Cell::new(true));
 
-        Log {
-            rawp: mem,
-            rawb: b,
-            size: num,
-            slog: raw,
-            head: CachePadded::new(AtomicUsize::new(0usize)),
-            tail: CachePadded::new(AtomicUsize::new(0usize)),
-            ctail: CachePadded::new(AtomicUsize::new(0usize)),
-            ltails: arr![Default::default(); 192],
-            next: CachePadded::new(AtomicUsize::new(1usize)),
-            lmasks: fls,
+        #[cfg(not(loom))]
+        {
+            #[allow(clippy::declare_interior_mutable_const)]
+            const LTAIL_DEFAULT: CachePadded<AtomicUsize> = CachePadded::new(AtomicUsize::new(0));
+
+            Log {
+                rawp: mem,
+                rawb: b,
+                size: num,
+                slog: raw,
+                head: CachePadded::new(AtomicUsize::new(0usize)),
+                tail: CachePadded::new(AtomicUsize::new(0usize)),
+                ctail: CachePadded::new(AtomicUsize::new(0usize)),
+                ltails: [LTAIL_DEFAULT; MAX_REPLICAS_PER_LOG],
+                next: CachePadded::new(AtomicUsize::new(1usize)),
+                lmasks: [LMASK_DEFAULT; MAX_REPLICAS_PER_LOG],
+            }
+        }
+        // AtomicUsize::new is not const in loom. This code block (including arr
+        // dependency) becomes redundant once
+        // https://github.com/tokio-rs/loom/issues/170 is fixed:
+        #[cfg(loom)]
+        {
+            use arr_macro::arr;
+            Log {
+                rawp: mem,
+                rawb: b,
+                size: num,
+                slog: raw,
+                head: CachePadded::new(AtomicUsize::new(0usize)),
+                tail: CachePadded::new(AtomicUsize::new(0usize)),
+                ctail: CachePadded::new(AtomicUsize::new(0usize)),
+                ltails: arr![CachePadded::new(AtomicUsize::new(0)); 3], // MAX_REPLICAS_PER_LOG
+                next: CachePadded::new(AtomicUsize::new(1usize)),
+                lmasks: [LMASK_DEFAULT; MAX_REPLICAS_PER_LOG],
+            }
         }
     }
 
@@ -251,7 +279,7 @@ where
     ///
     /// # Example
     ///
-    /// ```no-run
+    /// ```ignore
     /// use node_replication::Log;
     ///
     /// // Operation type that will go onto the log.
@@ -275,11 +303,15 @@ where
             let n = self.next.load(Ordering::Relaxed);
 
             // Check if we've exceeded the maximum number of replicas the log can support.
-            if n >= MAX_REPLICAS {
+            if n >= MAX_REPLICAS_PER_LOG {
                 return None;
             };
 
-            if self.next.compare_and_swap(n, n + 1, Ordering::SeqCst) != n {
+            if self
+                .next
+                .compare_exchange_weak(n, n + 1, Ordering::SeqCst, Ordering::SeqCst)
+                != Ok(n)
+            {
                 continue;
             };
 
@@ -291,7 +323,7 @@ where
     ///
     /// # Example
     ///
-    /// ```no-run
+    /// ```ignore
     /// use node_replication::Log;
     ///
     /// // Operation type that will go onto the log.
@@ -372,6 +404,9 @@ where
                 }
                 waitgc += 1;
                 self.exec(idx, &mut s);
+
+                #[cfg(loom)]
+                loom::thread::yield_now();
                 continue;
             }
 
@@ -384,16 +419,18 @@ where
 
             // Try reserving slots for the operations. If that fails, then restart
             // from the beginning of this loop.
-            if self
-                .tail
-                .compare_and_swap(tail, tail + nops, Ordering::Acquire)
-                != tail
+            if self.tail.compare_exchange_weak(
+                tail,
+                tail + nops,
+                Ordering::Acquire,
+                Ordering::Acquire,
+            ) != Ok(tail)
             {
                 continue;
             };
 
             // Successfully reserved entries on the shared log. Add the operations in.
-            for i in 0..nops {
+            for (i, op) in ops.iter().enumerate().take(nops) {
                 let e = self.slog[self.index(tail + i)].as_ptr();
                 let mut m = self.lmasks[idx - 1].get();
 
@@ -406,7 +443,7 @@ where
                     m = !m;
                 }
 
-                unsafe { (*e).operation = Some(ops[i].clone()) };
+                unsafe { (*e).operation = Some(op.clone()) };
                 unsafe { (*e).replica = idx };
                 unsafe { (*e).alivef.store(m, Ordering::Release) };
             }
@@ -426,7 +463,7 @@ where
     ///
     /// # Example
     ///
-    /// ```no-run
+    /// ```ignore
     /// use node_replication::Log;
     ///
     /// // Operation type that will go onto the log.
@@ -466,26 +503,26 @@ where
     #[inline(always)]
     pub(crate) fn exec<F: FnMut(T, usize)>(&self, idx: usize, d: &mut F) {
         // Load the logical log offset from which we must execute operations.
-        let l = self.ltails[idx - 1].load(Ordering::Relaxed);
+        let ltail = self.ltails[idx - 1].load(Ordering::Relaxed);
 
         // Check if we have any work to do by comparing our local tail with the log's
         // global tail. If they're equal, then we're done here and can simply return.
-        let t = self.tail.load(Ordering::Relaxed);
-        if l == t {
+        let gtail = self.tail.load(Ordering::Relaxed);
+        if ltail == gtail {
             return;
         }
 
         let h = self.head.load(Ordering::Relaxed);
 
         // Make sure we're within the shared log. If we aren't, then panic.
-        if l > t || l < h {
+        if ltail > gtail || ltail < h {
             panic!("Local tail not within the shared log!")
         };
 
         // Execute all operations from the passed in offset to the shared log's tail. Check if
         // the entry is live first; we could have a replica that has reserved entries, but not
         // filled them into the log yet.
-        for i in l..t {
+        for i in ltail..gtail {
             let mut iteration = 1;
             let e = self.slog[self.index(i)].as_ptr();
 
@@ -500,6 +537,9 @@ where
                     );
                 }
                 iteration += 1;
+
+                #[cfg(loom)]
+                loom::thread::yield_now();
             }
 
             unsafe { d((*e).operation.as_ref().unwrap().clone(), (*e).replica) };
@@ -513,8 +553,8 @@ where
 
         // Update the completed tail after we've executed these operations.
         // Also update this replica's local tail.
-        self.ctail.fetch_max(t, Ordering::Relaxed);
-        self.ltails[idx - 1].store(t, Ordering::Relaxed);
+        self.ctail.fetch_max(gtail, Ordering::Relaxed);
+        self.ltails[idx - 1].store(gtail, Ordering::Relaxed);
     }
 
     /// Returns a physical index given a logical index into the shared log.
@@ -556,6 +596,9 @@ where
                 }
                 iteration += 1;
                 self.exec(rid, &mut s);
+
+                #[cfg(loom)]
+                loom::thread::yield_now();
                 continue;
             }
 
@@ -591,7 +634,7 @@ where
         self.next.store(1, Ordering::SeqCst);
 
         // Next, reset replica-local metadata.
-        for r in 0..MAX_REPLICAS {
+        for r in 0..MAX_REPLICAS_PER_LOG {
             self.ltails[r].store(0, Ordering::Relaxed);
             self.lmasks[r].set(true);
         }
@@ -610,7 +653,7 @@ where
     ///
     /// # Example
     ///
-    /// ```no-run
+    /// ```ignore
     /// use node_replication::Log;
     ///
     /// // Operation type that will go onto the log.
@@ -751,11 +794,11 @@ mod tests {
         assert_eq!(l.next.load(Ordering::Relaxed), 1);
         assert_eq!(l.ctail.load(Ordering::Relaxed), 0);
 
-        for i in 0..MAX_REPLICAS {
+        for i in 0..MAX_REPLICAS_PER_LOG {
             assert_eq!(l.ltails[i].load(Ordering::Relaxed), 0);
         }
 
-        for i in 0..MAX_REPLICAS {
+        for i in 0..MAX_REPLICAS_PER_LOG {
             assert_eq!(l.lmasks[i].get(), true);
         }
     }
@@ -793,11 +836,11 @@ mod tests {
         assert_eq!(l.next.load(Ordering::Relaxed), 1);
         assert_eq!(l.ctail.load(Ordering::Relaxed), 0);
 
-        for i in 0..MAX_REPLICAS {
+        for i in 0..MAX_REPLICAS_PER_LOG {
             assert_eq!(l.ltails[i].load(Ordering::Relaxed), 0);
         }
 
-        for i in 0..MAX_REPLICAS {
+        for i in 0..MAX_REPLICAS_PER_LOG {
             assert_eq!(l.lmasks[i].get(), true);
         }
     }
@@ -821,9 +864,9 @@ mod tests {
     #[test]
     fn test_log_register_none() {
         let l = Log::<Operation>::new(1024);
-        l.next.store(MAX_REPLICAS, Ordering::Relaxed);
+        l.next.store(MAX_REPLICAS_PER_LOG, Ordering::Relaxed);
         assert!(l.register().is_none());
-        assert_eq!(l.next.load(Ordering::Relaxed), MAX_REPLICAS);
+        assert_eq!(l.next.load(Ordering::Relaxed), MAX_REPLICAS_PER_LOG);
     }
 
     // Test that we can correctly append an entry into the log.
