@@ -2,13 +2,19 @@
 // SPDX-License-Identifier: Apache-2.0 OR MIT
 
 use core::cell::RefCell;
-use core::mem::MaybeUninit;
-use core::sync::atomic::{spin_loop_hint, AtomicUsize, Ordering};
+use core::hint::spin_loop;
+#[cfg(not(loom))]
+use core::sync::atomic::{AtomicUsize, Ordering};
+#[cfg(loom)]
+use loom::sync::atomic::{AtomicUsize, Ordering};
 
+#[cfg(not(loom))]
 use alloc::sync::Arc;
+#[cfg(loom)]
+use loom::sync::Arc;
+
 use alloc::vec::Vec;
 
-use arr_macro::arr;
 use crossbeam_utils::CachePadded;
 
 use super::context::Context;
@@ -26,6 +32,7 @@ use super::Dispatch;
 pub struct ReplicaToken(usize);
 
 /// To make it harder to use the same ReplicaToken on multiple threads.
+#[cfg(features = "unstable")]
 impl !Send for ReplicaToken {}
 
 impl ReplicaToken {
@@ -52,7 +59,10 @@ impl ReplicaToken {
 /// # Important
 /// If this number is adjusted due to the use of the `arr_macro::arr` macro we
 /// have to adjust the `256` literals in the `new` constructor of `Replica`.
+#[cfg(not(loom))]
 pub const MAX_THREADS_PER_REPLICA: usize = 256;
+#[cfg(loom)]
+pub const MAX_THREADS_PER_REPLICA: usize = 2;
 const_assert!(
     MAX_THREADS_PER_REPLICA >= 1 && (MAX_THREADS_PER_REPLICA & (MAX_THREADS_PER_REPLICA - 1) == 0)
 );
@@ -70,7 +80,7 @@ const_assert!(
 /// the same underlying log.
 pub struct Replica<'a, D>
 where
-    D: Sized + Default + Dispatch + Sync,
+    D: Sized + Dispatch + Sync,
 {
     /// A replica-identifier received when the replica is registered against
     /// the shared-log. Required when consuming operations from the log.
@@ -115,11 +125,11 @@ where
 
 /// The Replica is Sync. Member variables are protected by a CAS on `combiner`.
 /// Contexts are thread-safe.
-unsafe impl<'a, D> Sync for Replica<'a, D> where D: Sized + Default + Sync + Dispatch {}
+unsafe impl<'a, D> Sync for Replica<'a, D> where D: Sized + Sync + Dispatch {}
 
 impl<'a, D> core::fmt::Debug for Replica<'a, D>
 where
-    D: Sized + Default + Sync + Dispatch,
+    D: Sized + Sync + Dispatch,
 {
     fn fmt(&self, f: &mut core::fmt::Formatter) -> core::fmt::Result {
         write!(f, "Replica")
@@ -181,10 +191,78 @@ where
     /// let replica = Replica::<Data>::new(&log);
     /// ```
     pub fn new<'b>(log: &Arc<Log<'b, <D as Dispatch>::WriteOperation>>) -> Arc<Replica<'b, D>> {
+        Replica::with_data(log, Default::default())
+    }
+}
+
+impl<'a, D> Replica<'a, D>
+where
+    D: Sized + Dispatch + Sync,
+{
+    /// Similar to [`Replica<D>::new`], but we pass a pre-initialized
+    /// data-structure as an argument (`d`) rather than relying on the
+    /// [Default](core::default::Default) trait to create one.
+    ///
+    /// # Note
+    /// [`Replica<D>::new`] should be the preferred method to create a Replica.
+    /// If `with_data` is used, care must be taken that the same state is passed
+    /// to every Replica object. If not the resulting operations executed
+    /// against replicas may not give deterministic results.
+    #[cfg(not(feature = "unstable"))]
+    pub fn with_data<'b>(
+        log: &Arc<Log<'b, <D as Dispatch>::WriteOperation>>,
+        d: D,
+    ) -> Arc<Replica<'b, D>> {
+        let mut contexts = Vec::with_capacity(MAX_THREADS_PER_REPLICA);
+        // Add `MAX_THREADS_PER_REPLICA` contexts
+        for _idx in 0..MAX_THREADS_PER_REPLICA {
+            contexts.push(Default::default());
+        }
+
+        Arc::new(
+            Replica {
+                idx: log.register().unwrap(),
+                combiner: CachePadded::new(AtomicUsize::new(0)),
+                next: CachePadded::new(AtomicUsize::new(1)),
+                contexts,
+                buffer:
+                    RefCell::new(
+                        Vec::with_capacity(
+                            MAX_THREADS_PER_REPLICA
+                                * Context::<
+                                    <D as Dispatch>::WriteOperation,
+                                    <D as Dispatch>::Response,
+                                >::batch_size(),
+                        ),
+                    ),
+                inflight: RefCell::new([0; MAX_THREADS_PER_REPLICA]),
+                result:
+                    RefCell::new(
+                        Vec::with_capacity(
+                            MAX_THREADS_PER_REPLICA
+                                * Context::<
+                                    <D as Dispatch>::WriteOperation,
+                                    <D as Dispatch>::Response,
+                                >::batch_size(),
+                        ),
+                    ),
+                slog: log.clone(),
+                data: CachePadded::new(RwLock::<D>::new(d)),
+            },
+        )
+    }
+
+    /// See `with_data` documentation without unstable feature.
+    #[cfg(feature = "unstable")]
+    pub fn with_data<'b>(
+        log: &Arc<Log<'b, <D as Dispatch>::WriteOperation>>,
+        d: D,
+    ) -> Arc<Replica<'b, D>> {
+        use core::mem::MaybeUninit;
         let mut uninit_replica: Arc<MaybeUninit<Replica<D>>> = Arc::new_zeroed();
 
-        // This is the preferred but unsafe mode of initialization as it avoids
-        // putting the (often big) Replica object on the stack first.
+        // This is the preferred (but unsafe) mode of initialization as it avoids
+        // putting the big Replica object on the stack first.
         unsafe {
             let uninit_ptr = Arc::get_mut_unchecked(&mut uninit_replica).as_mut_ptr();
             uninit_ptr.write(Replica {
@@ -202,7 +280,7 @@ where
                                 >::batch_size(),
                         ),
                     ),
-                inflight: RefCell::new(arr![Default::default(); 256]),
+                inflight: RefCell::new([0; MAX_THREADS_PER_REPLICA]),
                 result:
                     RefCell::new(
                         Vec::with_capacity(
@@ -214,7 +292,7 @@ where
                         ),
                     ),
                 slog: log.clone(),
-                data: CachePadded::new(RwLock::<D>::default()),
+                data: CachePadded::new(RwLock::<D>::new(d)),
             });
 
             let mut replica = uninit_replica.assume_init();
@@ -284,7 +362,11 @@ where
                 return None;
             };
 
-            if self.next.compare_and_swap(idx, idx + 1, Ordering::SeqCst) != idx {
+            if self
+                .next
+                .compare_exchange_weak(idx, idx + 1, Ordering::SeqCst, Ordering::SeqCst)
+                != Ok(idx)
+            {
                 continue;
             };
 
@@ -414,8 +496,8 @@ where
         // times with no luck, try to perform flat combining to make some progress.
         loop {
             let r = self.contexts[idx - 1].res();
-            if r.is_some() {
-                return r.unwrap();
+            if let Some(resp) = r {
+                return resp;
             }
 
             iter += 1;
@@ -438,14 +520,17 @@ where
     pub fn verify<F: FnMut(&D)>(&self, mut v: F) {
         // Acquire the combiner lock before attempting anything on the data structure.
         // Use an idx greater than the maximum that can be allocated.
-        while self
-            .combiner
-            .compare_and_swap(0, MAX_THREADS_PER_REPLICA + 2, Ordering::Acquire)
-            != 0
-        {}
+        while self.combiner.compare_exchange_weak(
+            0,
+            MAX_THREADS_PER_REPLICA + 2,
+            Ordering::Acquire,
+            Ordering::Acquire,
+        ) != Ok(0)
+        {
+            spin_loop();
+        }
 
         let mut data = self.data.write(self.next.load(Ordering::Relaxed));
-
         let mut f = |o: <D as Dispatch>::WriteOperation, _i: usize| {
             data.dispatch_mut(o);
         };
@@ -465,7 +550,7 @@ where
         let ctail = self.slog.get_ctail();
         while !self.slog.is_replica_synced_for_reads(self.idx, ctail) {
             self.try_combine(idx.0);
-            spin_loop_hint();
+            spin_loop();
         }
     }
 
@@ -481,10 +566,10 @@ where
         let ctail = self.slog.get_ctail();
         while !self.slog.is_replica_synced_for_reads(self.idx, ctail) {
             self.try_combine(tid);
-            spin_loop_hint();
+            spin_loop();
         }
 
-        self.data.read(tid - 1).dispatch(op)
+        return self.data.read(tid - 1).dispatch(op);
     }
 
     /// Enqueues an operation inside a thread local context. Returns a boolean
@@ -500,6 +585,7 @@ where
         // First, check if there already is a flat combiner. If there is no active flat combiner
         // then try to acquire the combiner lock. If there is, then just return.
         for _i in 0..4 {
+            #[cfg(not(loom))]
             if unsafe {
                 core::ptr::read_volatile(
                     &self.combiner
@@ -509,11 +595,25 @@ where
             } != 0
             {
                 return;
-            };
+            }
+
+            #[cfg(loom)]
+            {
+                if self.combiner.load(Ordering::Relaxed) != 0 {
+                    loom::thread::yield_now();
+                    return;
+                }
+            }
         }
 
         // Try to become the combiner here. If this fails, then simply return.
-        if self.combiner.compare_and_swap(0, tid, Ordering::Acquire) != 0 {
+        if self
+            .combiner
+            .compare_exchange_weak(0, tid, Ordering::Acquire, Ordering::Acquire)
+            != Ok(0)
+        {
+            #[cfg(loom)]
+            loom::thread::yield_now();
             return;
         }
 
@@ -546,8 +646,12 @@ where
         // Append all collected operations into the shared log. We pass a closure
         // in here because operations on the log might need to be consumed for GC.
         {
+            let mut data = self.data.write(next);
             let f = |o: <D as Dispatch>::WriteOperation, i: usize| {
-                let resp = self.data.write(next).dispatch_mut(o);
+                #[cfg(not(loom))]
+                let resp = data.dispatch_mut(o);
+                #[cfg(loom)]
+                let resp = data.dispatch_mut(o);
                 if i == self.idx {
                     results.push(resp);
                 }
