@@ -6,11 +6,19 @@
 //! This module is only public since it needs to be exposed to the benchmarking
 //! code. For clients there is no need to rely on this directly, as the RwLock
 //! is embedded inside the Replica.
+//!
+//! # Testing with loom
+//!
+//! We're not using loom in this module because we use UnsafeCell and loom's
+//! UnsafeCell exposes a different API. Luckily, loom provides it's own RwLock
+//! implementation which (with some modifications, see `loom_rwlock.rs`) we can
+//! use in the replica code.
 
 use core::cell::UnsafeCell;
 use core::default::Default;
+use core::hint::spin_loop;
 use core::ops::{Deref, DerefMut};
-use core::sync::atomic::{spin_loop_hint, AtomicBool, AtomicUsize, Ordering};
+use core::sync::atomic::{AtomicBool, AtomicUsize, Ordering};
 
 use crossbeam_utils::CachePadded;
 
@@ -18,6 +26,9 @@ use crossbeam_utils::CachePadded;
 const MAX_READER_THREADS: usize = 192;
 const_assert!(MAX_READER_THREADS > 0);
 
+#[allow(clippy::declare_interior_mutable_const)]
+const RLOCK_DEFAULT: CachePadded<AtomicUsize> = CachePadded::new(AtomicUsize::new(0));
+
 /// A scalable reader-writer lock.
 ///
 /// This lock favours reader performance over writers. Each reader thread gets
@@ -28,7 +39,7 @@ const_assert!(MAX_READER_THREADS > 0);
 /// Calling `write()` returns a write-guard that can be used to safely mutate `T`.
 pub struct RwLock<T>
 where
-    T: Sized + Default + Sync,
+    T: Sized + Sync,
 {
     /// The writer lock. There can be at most one writer at any given point of time.
     wlock: CachePadded<AtomicBool>,
@@ -42,7 +53,7 @@ where
 
 /// A read-guard that can be used to read the underlying data structure. Writes on
 /// the data structure will be blocked as long as one of these is lying around.
-pub struct ReadGuard<'a, T: ?Sized + Default + Sync + 'a> {
+pub struct ReadGuard<'a, T: Sized + Sync + 'a> {
     /// Id of the thread that acquired this guard. Required at drop time so that
     /// we can release the appropriate read lock.
     tid: usize,
@@ -53,7 +64,7 @@ pub struct ReadGuard<'a, T: ?Sized + Default + Sync + 'a> {
 
 /// A write-guard that can be used to write to the underlying data structure. All
 /// reads will be blocked until this is dropped.
-pub struct WriteGuard<'a, T: ?Sized + Default + Sync + 'a> {
+pub struct WriteGuard<'a, T: Sized + Sync + 'a> {
     /// A reference to the Rwlock wrapping the data-structure.
     lock: &'a RwLock<T>,
 }
@@ -65,11 +76,9 @@ where
     /// Returns a new instance of a RwLock. Default constructs the
     /// underlying data structure.
     fn default() -> RwLock<T> {
-        use arr_macro::arr;
-
         RwLock {
             wlock: CachePadded::new(AtomicBool::new(false)),
-            rlock: arr![Default::default(); 192],
+            rlock: [RLOCK_DEFAULT; MAX_READER_THREADS],
             data: UnsafeCell::new(T::default()),
         }
     }
@@ -77,8 +86,18 @@ where
 
 impl<T> RwLock<T>
 where
-    T: Sized + Default + Sync,
+    T: Sized + Sync,
 {
+    /// Returns a new instance of a RwLock. Default constructs the
+    /// underlying data structure.
+    pub fn new(t: T) -> Self {
+        Self {
+            wlock: CachePadded::new(AtomicBool::new(false)),
+            rlock: [RLOCK_DEFAULT; MAX_READER_THREADS],
+            data: UnsafeCell::new(t),
+        }
+    }
+
     /// Locks the underlying data-structure for writes. The caller can retrieve
     /// a mutable reference from the returned `WriteGuard`.
     ///
@@ -101,8 +120,16 @@ where
     /// ```
     pub fn write(&self, n: usize) -> WriteGuard<T> {
         // First, wait until we can acquire the writer lock.
-        while self.wlock.compare_and_swap(false, true, Ordering::Acquire) {
-            spin_loop_hint();
+        loop {
+            match self.wlock.compare_exchange_weak(
+                false,
+                true,
+                Ordering::Acquire,
+                Ordering::Acquire,
+            ) {
+                Ok(_) => break,
+                Err(_) => continue,
+            }
         }
 
         // Next, wait until all readers have released their locks. This condition
@@ -113,7 +140,7 @@ where
             .take(n)
             .all(|item| item.load(Ordering::Relaxed) == 0)
         {
-            spin_loop_hint();
+            spin_loop();
         }
 
         unsafe { WriteGuard::new(self) }
@@ -150,7 +177,7 @@ where
             // optimization spoken of earlier.
             unsafe {
                 while core::ptr::read_volatile(ptr) {
-                    spin_loop_hint();
+                    spin_loop();
                 }
             }
 
@@ -171,8 +198,12 @@ where
 
     /// Unlocks the write lock; invoked by the drop() method.
     pub(in crate::rwlock) unsafe fn write_unlock(&self) {
-        if !self.wlock.compare_and_swap(true, false, Ordering::Acquire) {
-            panic!("write_unlock() called without acquiring the write lock");
+        match self
+            .wlock
+            .compare_exchange_weak(true, false, Ordering::Acquire, Ordering::Acquire)
+        {
+            Ok(_) => (),
+            Err(_) => panic!("write_unlock() called without acquiring the write lock"),
         }
     }
 
@@ -184,14 +215,14 @@ where
     }
 }
 
-impl<'rwlock, T: ?Sized + Default + Sync> ReadGuard<'rwlock, T> {
+impl<'rwlock, T: Sized + Sync> ReadGuard<'rwlock, T> {
     /// Returns a read guard over a passed in reader-writer lock.
     unsafe fn new(lock: &'rwlock RwLock<T>, tid: usize) -> ReadGuard<'rwlock, T> {
         ReadGuard { tid, lock }
     }
 }
 
-impl<'rwlock, T: ?Sized + Default + Sync> WriteGuard<'rwlock, T> {
+impl<'rwlock, T: Sized + Sync> WriteGuard<'rwlock, T> {
     /// Returns a write guard over a passed in reader-writer lock.
     unsafe fn new(lock: &'rwlock RwLock<T>) -> WriteGuard<'rwlock, T> {
         WriteGuard { lock }
@@ -201,11 +232,11 @@ impl<'rwlock, T: ?Sized + Default + Sync> WriteGuard<'rwlock, T> {
 /// `Sync` trait allows `RwLock` to be shared between threads. The `read()` and
 /// `write()` logic ensures that we will never have threads writing to and
 /// reading from the underlying data structure simultaneously.
-unsafe impl<T: ?Sized + Default + Sync> Sync for RwLock<T> {}
+unsafe impl<T: Sized + Sync> Sync for RwLock<T> {}
 
 /// This `Deref` trait allows a thread to use T from a ReadGuard.
 /// ReadGuard can only be dereferenced into an immutable reference.
-impl<T: ?Sized + Default + Sync> Deref for ReadGuard<'_, T> {
+impl<T: Sized + Sync> Deref for ReadGuard<'_, T> {
     type Target = T;
 
     fn deref(&self) -> &T {
@@ -215,7 +246,7 @@ impl<T: ?Sized + Default + Sync> Deref for ReadGuard<'_, T> {
 
 /// This `Deref` trait allows a thread to use T from a WriteGuard.
 /// This allows us to dereference an immutable reference.
-impl<T: ?Sized + Default + Sync> Deref for WriteGuard<'_, T> {
+impl<T: Sized + Sync> Deref for WriteGuard<'_, T> {
     type Target = T;
 
     fn deref(&self) -> &T {
@@ -225,7 +256,7 @@ impl<T: ?Sized + Default + Sync> Deref for WriteGuard<'_, T> {
 
 /// This `DerefMut` trait allow a thread to use T from a WriteGuard.
 /// This allows us to dereference a mutable reference.
-impl<T: ?Sized + Default + Sync> DerefMut for WriteGuard<'_, T> {
+impl<T: Sized + Sync> DerefMut for WriteGuard<'_, T> {
     fn deref_mut(&mut self) -> &mut T {
         unsafe { &mut *self.lock.data.get() }
     }
@@ -233,7 +264,7 @@ impl<T: ?Sized + Default + Sync> DerefMut for WriteGuard<'_, T> {
 
 /// This `Drop` trait implements the unlock logic for a reader lock. Once the `ReadGuard`
 /// goes out of scope, the corresponding read lock is marked as released.
-impl<T: ?Sized + Default + Sync> Drop for ReadGuard<'_, T> {
+impl<T: Sized + Sync> Drop for ReadGuard<'_, T> {
     fn drop(&mut self) {
         unsafe {
             let tid = self.tid;
@@ -244,7 +275,7 @@ impl<T: ?Sized + Default + Sync> Drop for ReadGuard<'_, T> {
 
 /// This `Drop` trait implements the unlock logic for a writer lock. Once the `WriteGuard`
 /// goes out of scope, the corresponding write lock is marked as released.
-impl<T: ?Sized + Default + Sync> Drop for WriteGuard<'_, T> {
+impl<T: Sized + Sync> Drop for WriteGuard<'_, T> {
     fn drop(&mut self) {
         unsafe {
             self.lock.write_unlock();
